#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module convert pandas's data type into varieties of form.
"""

import sys
if sys.version_info[0] == 2:
    from itertools import (
        izip as zip,
    )
import json
import base64
from collections import OrderedDict
import pandas as pd
try:
    from .pkg.rolex import rolex
except:
    from pandas_mate.pkg.rolex import rolex


def itertuple(df):
    """High performance tuple style iterator.

    :param df: ``pandas.DataFrame`` instance.

    **中文文档**

    对 ``pandas.DataFrame`` 进行tuple风格的高性能行遍历。
    """
    return zip(*(l for col, l in df.iteritems()))


def grouper_df(df, chunksize):
    """Evenly divide pd.DataFrame into n rows piece, no filled value 
    if sub dataframe's size smaller than n.

    :param df: ``pandas.DataFrame`` instance.
    :param chunksize: number of rows of each small DataFrame.

    **中文文档**

    将 ``pandas.DataFrame`` 分拆成等大小的小DataFrame。
    """
    data = list()
    counter = 0
    for tp in zip(*(l for col, l in df.iteritems())):
        counter += 1
        data.append(tp)
        if counter == chunksize:
            new_df = pd.DataFrame(data, columns=df.columns)
            yield new_df
            data = list()
            counter = 0

    if len(data) > 0:
        new_df = pd.DataFrame(data, columns=df.columns)
        yield new_df


def to_index_row_dict(df, index_col=None, use_ordered_dict=True):
    """Transform data frame to list of dict.

    :param index_col: None or str, the column that used as index.
    :param use_ordered_dict: if True, row dict is has same order as df.columns. 

    **中文文档**

    将dataframe以指定列为key, 转化成以行为视角的dict结构, 提升按行index访问
    的速度。若无指定列, 则使用index。
    """
    if index_col:
        index_list = df[index_col]
    else:
        index_list = df.index

    columns = df.columns

    if use_ordered_dict:
        table = OrderedDict()
    else:
        table = dict()

    for ind, tp in zip(index_list, itertuple(df)):
        table[ind] = dict(zip(columns, tp))

    return table


def to_dict_list(df, use_ordered_dict=True):
    """Transform each row to dict, and put them into a list.

    **中文文档**

    将 ``pandas.DataFrame`` 转换成一个字典的列表。列表的长度与行数相同, 其中
    每一个字典相当于表中的一行, 相当于一个 ``pandas.Series`` 对象。
    """
    if use_ordered_dict:
        dict = OrderedDict

    columns = df.columns
    data = list()
    for tp in itertuple(df):
        data.append(dict(zip(columns, tp)))
    return data


def to_dict_list_generic_type(df, int_col=None, binary_col=None):
    """Transform each row to dict, and put them into a list. And automatically
    convert ``np.int64`` to ``int``, ``pandas.tslib.Timestamp`` to 
    ``datetime.datetime``, ``np.nan`` to ``None``.

    :param df: ``pandas.DataFrame`` instance.
    :param int_col: integer type columns.
    :param binary_col: binary type type columns.

    **中文文档**

    由于 ``pandas.Series`` 中的值的整数数据类型是 ``numpy.int64``, 
    时间数据类型是 ``pandas.tslib.Timestamp``, None的数据类型是 ``np.nan``。
    虽然从访问和计算的角度来说没有什么问题, 但会和很多数据库的操作不兼容。 

    此函数能将 ``pandas.DataFrame`` 转化成字典的列表。数据类型能正确的获得int, 
    bytes和datetime.datetime。
    """
    # Pre-process int_col, binary_col and datetime_col
    if (int_col is not None) and (not isinstance(int_col, (list, tuple))):
        int_col = [int_col, ]

    if (binary_col is not None) and (not isinstance(binary_col, (list, tuple))):
        binary_col = [binary_col, ]

    datetime_col = list()
    for col, dtype in dict(df.dtypes).items():
        if "datetime64" in str(dtype):
            datetime_col.append(col)
    if len(datetime_col) == 0:
        datetime_col = None

    # Pre-process binary column dataframe
    def b64_encode(b):
        try:
            return base64.b64encode(b)
        except:
            return b

    if binary_col is not None:
        for col in binary_col:
            df[col] = df[col].apply(b64_encode)

    data = json.loads(df.to_json(orient="records", date_format="iso"))

    if int_col is not None:
        for row in data:
            for col in int_col:
                try:
                    row[col] = int(row[col])
                except:
                    pass

    if binary_col is not None:
        for row in data:
            for col in binary_col:
                try:
                    row[col] = base64.b64decode(row[col].encode("ascii"))
                except:
                    pass

    if datetime_col is not None:
        for row in data:
            for col in datetime_col:
                try:
                    row[col] = rolex.str2datetime(row[col])
                except:
                    pass

    return data



"""
Table factories
"""
import csv
from .prettytable import PrettyTable
from ._compact import py3k, HTMLParser


def from_csv(fp, field_names=None, **kwargs):
    fmtparams = {}
    for param in ["delimiter", "doublequote", "escapechar", "lineterminator",
                  "quotechar", "quoting", "skipinitialspace", "strict"]:
        if param in kwargs:
            fmtparams[param] = kwargs.pop(param)
    if fmtparams:
        reader = csv.reader(fp, **fmtparams)
    else:
        dialect = csv.Sniffer().sniff(fp.read(1024))
        fp.seek(0)
        reader = csv.reader(fp, dialect)

    table = PrettyTable(**kwargs)
    if field_names:
        table.field_names = field_names
    else:
        if py3k:
            table.field_names = [x.strip() for x in next(reader)]
        else:
            table.field_names = [x.strip() for x in reader.next()]

    for row in reader:
        table.add_row([x.strip() for x in row])

    return table


def from_db_cursor(cursor, **kwargs):
    if cursor.description:
        table = PrettyTable(**kwargs)
        table.field_names = [col[0] for col in cursor.description]
        for row in cursor.fetchall():
            table.add_row(row)
        return table


class TableHandler(HTMLParser):
    def __init__(self, **kwargs):
        HTMLParser.__init__(self)
        self.kwargs = kwargs
        self.tables = []
        self.last_row = []
        self.rows = []
        self.max_row_width = 0
        self.active = None
        self.last_content = ""
        self.is_last_row_header = False
        self.colspan = 0

    def handle_starttag(self, tag, attrs):
        self.active = tag
        if tag == "th":
            self.is_last_row_header = True
        for (key, value) in attrs:
            if key == "colspan":
                self.colspan = int(value)

    def handle_endtag(self, tag):
        if tag in ["th", "td"]:
            stripped_content = self.last_content.strip()
            self.last_row.append(stripped_content)
            if self.colspan:
                for i in range(1, self.colspan):
                    self.last_row.append("")
                self.colspan = 0

        if tag == "tr":
            self.rows.append(
                (self.last_row, self.is_last_row_header))
            self.max_row_width = max(self.max_row_width, len(self.last_row))
            self.last_row = []
            self.is_last_row_header = False
        if tag == "table":
            table = self.generate_table(self.rows)
            self.tables.append(table)
            self.rows = []
        self.last_content = " "
        self.active = None

    def handle_data(self, data):
        self.last_content += data

    def generate_table(self, rows):
        """
        Generates from a list of rows a PrettyTable object.
        """
        table = PrettyTable(**self.kwargs)
        for row in self.rows:
            if len(row[0]) < self.max_row_width:
                appends = self.max_row_width - len(row[0])
                for i in range(1, appends):
                    row[0].append("-")

            if row[1] is True:
                self.make_fields_unique(row[0])
                table.field_names = row[0]
            else:
                table.add_row(row[0])
        return table

    def make_fields_unique(self, fields):
        """
        iterates over the row and make each field unique
        """
        for i in range(0, len(fields)):
            for j in range(i + 1, len(fields)):
                if fields[i] == fields[j]:
                    fields[j] += "'"


def from_html(html_code, **kwargs):
    """
    Generates a list of PrettyTables from a string of HTML code. Each <table> in
    the HTML becomes one PrettyTable object.
    """

    parser = TableHandler(**kwargs)
    parser.feed(html_code)
    return parser.tables


def from_html_one(html_code, **kwargs):
    """
    Generates a PrettyTables from a string of HTML code which contains only a
    single <table>
    """

    tables = from_html(html_code, **kwargs)
    try:
        assert len(tables) == 1
    except AssertionError:
        raise Exception(
            "More than one <table> in provided HTML code!  Use from_html instead.")
    return tables[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module provide database I/O interface.
"""

import math
import pandas as pd
try:
    from .transform import grouper_df, to_dict_list_generic_type
except:
    from pandas_mate.transform import grouper_df, to_dict_list_generic_type


def smart_insert(df, table, engine, minimal_size=5):
    """An optimized Insert strategy.

    **中文文档**

    一种优化的将大型DataFrame中的数据, 在有IntegrityError的情况下将所有
    好数据存入数据库的方法。
    """
    from sqlalchemy.exc import IntegrityError

    try:
        table_name = table.name
    except:
        table_name = table

    # 首先进行尝试bulk insert
    try:
        df.to_sql(table_name, engine, index=False, if_exists="append")
    # 失败了
    except IntegrityError:
        # 分析数据量
        n = df.shape[0]
        # 如果数据条数多于一定数量
        if n >= minimal_size ** 2:
            # 则进行分包
            n_chunk = math.floor(math.sqrt(n))
            for sub_df in grouper_df(df, n_chunk):
                smart_insert(
                    sub_df, table_name, engine, minimal_size)
        # 否则则一条条地逐条插入
        else:
            for sub_df in grouper_df(df, 1):
                try:
                    sub_df.to_sql(
                        table_name, engine, index=False, if_exists="append")
                except IntegrityError:
                    pass


def excel_to_sql(excel_file_path, engine,
                 read_excel_kwargs=None,
                 to_generic_type_kwargs=None,
                 to_sql_kwargs=None):
    """Create a database from excel.

    :param read_excel_kwargs: dict, arguments for ``pandas.read_excel`` method.
      example: ``{"employee": {"skiprows": 10}, "department": {}}``
    :param to_sql_kwargs: dict, arguments for ``pandas.DataFrame.to_sql`` 
      method.

    limitation:

    1. If a integer column has None value, data type in database will be float.
      Because pandas thinks that it is ``np.nan``.
    2. If a string column looks like integer, ``pandas.read_excel()`` method
      doesn't have options to convert it to string.
    """
    if read_excel_kwargs is None:
        read_excel_kwargs = dict()

    if to_sql_kwargs is None:
        to_sql_kwargs = dict()

    if to_generic_type_kwargs is None:
        to_generic_type_kwargs = dict()

    xl = pd.ExcelFile(excel_file_path)
    for sheet_name in xl.sheet_names:
        df = pd.read_excel(
            excel_file_path, sheet_name,
            **read_excel_kwargs.get(sheet_name, dict())
        )

        kwargs = to_generic_type_kwargs.get(sheet_name)
        if kwargs:
            data = to_dict_list_generic_type(df, **kwargs)
            smart_insert(data, sheet_name, engine)
        else:
            df.to_sql(
                sheet_name, engine, index=False,
                **to_sql_kwargs.get(sheet_name, dict(if_exists="replace"))
            )


def database_to_excel(engine, excel_file_path):
    """Export database to excel.

    :param engine: 
    :param excel_file_path:
    """
    from sqlalchemy import MetaData, select

    metadata = MetaData()
    metadata.reflect(engine)

    writer = pd.ExcelWriter(excel_file_path)
    for table in metadata.tables.values():
        sql = select([table])
        df = pd.read_sql(sql, engine)
        df.to_excel(writer, table.name, index=False)

    writer.save()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from superjson import json

try:
    from .pkg.sixmini import iteritems
except:  # pragma: no cover
    from pymongo_mate.pkg.sixmini import iteritems


def _dump(db):
    """
    Dump :class:`mongomock.database.Database` to dict data.
    """
    db_data = {"name": db.name, "_collections": dict()}

    for col_name, collection in iteritems(db._collections):
        if col_name != "system.indexes":
            col_data = {
                "_documents": collection._documents,
                "_uniques": collection._uniques,
            }
            db_data["_collections"][col_name] = col_data

    return db_data


def _load(db_data, db):
    """
    Load :class:`mongomock.database.Database` from dict data.
    """
    if db.name != db_data["name"]:
        raise ValueError("dbname doesn't matches! Maybe wrong database data.")

    db.__init__(client=db._client, name=db.name)
    for col_name, col_data in iteritems(db_data["_collections"]):
        collection = db.get_collection(col_name)
        collection._documents = col_data["_documents"]
        collection._uniques = col_data["_uniques"]
        db._collections[col_name] = collection

    return db


def dump_db(db, file,
            pretty=False,
            overwrite=False,
            verbose=True):
    """
    Dump :class:`mongomock.database.Database` to a local file. Only support
    ``*.json`` or ``*.gz`` (compressed json file)

    :param db: instance of :class:`mongomock.database.Database`.
    :param file: file path.
    :param pretty: bool, toggle on jsonize into pretty format.
    :param overwrite: bool, allow overwrite to existing file.
    :param verbose: bool, toggle on log.
    """
    db_data = _dump(db)
    json.dump(
        db_data, file,
        pretty=pretty, overwrite=overwrite, verbose=verbose,
    )


def load_db(file, db, verbose=True):
    """
    Load :class:`mongomock.database.Database` from a local file.

    :param file: file path.
    :param db: instance of :class:`mongomock.database.Database`.
    :param verbose: bool, toggle on log.
    :return: loaded db.
    """
    db_data = json.load(file, verbose=verbose)
    return _load(db_data, db)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2014-2016 Sanhe Hu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import print_function, unicode_literals
import os
import site
import shutil
import hashlib
import platform

__version__ = "0.0.1"
__short_description__ = "Utility script to install your package in one shot, for developer."
__author__ = "Sanhe Hu"

SRC = os.getcwd()
PKG_NAME = os.path.basename(SRC)

SYS_NAME = platform.system()
if SYS_NAME == "Windows":
    BIN_SCRIPTS = "Scripts"
elif SYS_NAME in ["Darwin", "Linux"]:
    BIN_SCRIPTS = "bin"


def is_venv():
    """Check whether if this workspace is a virtualenv.
    """
    dir_path = os.path.dirname(SRC)
    is_venv_flag = True

    if SYS_NAME == "Windows":
        executable_list = ["activate", "pip.exe", "python.exe"]
    elif SYS_NAME in ["Darwin", "Linux"]:
        executable_list = ["activate", "pip", "python"]

    for executable in executable_list:
        path = os.path.join(dir_path, BIN_SCRIPTS, executable)
        if not os.path.exists(path):
            is_venv_flag = False

    return is_venv_flag


def find_linux_venv_py_version():
    """Find python version name used in this virtualenv.

    For example: ``python2.7``, ``python3.4``
    """
    available_python_version = [
        "python2.6",
        "python2.7",
        "python3.3",
        "python3.4",
        "python3.5",
        "python3.6",
    ]
    dir_path = os.path.dirname(SRC)
    for basename in os.listdir(os.path.join(dir_path, BIN_SCRIPTS)):
        for python_version in available_python_version:
            if python_version in basename:
                return python_version
    raise Exception("Can't find virtualenv python version!")


def find_venv_DST():
    """Find where this package should be installed to in this virtualenv.

    For example: ``/path-to-venv/lib/python2.7/site-packages/package-name``
    """
    dir_path = os.path.dirname(SRC)

    if SYS_NAME == "Windows":
        DST = os.path.join(dir_path, "Lib", "site-packages", PKG_NAME)
    elif SYS_NAME in ["Darwin", "Linux"]:
        python_version = find_linux_venv_py_version()
        DST = os.path.join(dir_path, "lib", python_version,
                           "site-packages", PKG_NAME)

    return DST


def find_DST():
    """Find where this package should be installed to.
    """
    if SYS_NAME == "Windows":
        return os.path.join(site.getsitepackages()[1], PKG_NAME)
    elif SYS_NAME in ["Darwin", "Linux"]:
        return os.path.join(site.getsitepackages()[0], PKG_NAME)


if is_venv():
    DST = find_venv_DST()
else:
    DST = find_DST()


def md5_of_file(abspath):
    """Md5 value of a file.
    """
    chunk_size = 1024 * 1024
    m = hashlib.md5()
    with open(abspath, "rb") as f:
        while True:
            data = f.read(chunk_size)
            if not data:
                break
            m.update(data)
    return m.hexdigest()


def check_need_install():
    """Check if installed package are exactly the same to this one.
    By checking md5 value of all files.
    """
    need_install_flag = False
    for root, _, basename_list in os.walk(SRC):
        if os.path.basename(root) != "__pycache__":
            for basename in basename_list:
                src = os.path.join(root, basename)
                dst = os.path.join(root.replace(SRC, DST), basename)
                if os.path.exists(dst):
                    if md5_of_file(src) != md5_of_file(dst):
                        return True
                else:
                    return True
    return need_install_flag


def install():
    """Manual install main script.
    """
    # check installed package
    print("Compare to '%s' ..." % DST)
    need_install_flag = check_need_install()
    if not need_install_flag:
        print("    package is up-to-date, no need to install.")
        return
    print("Difference been found, start installing ...")

    # remove __pycache__ folder and *.pyc file
    print("Remove *.pyc file ...")
    pyc_folder_list = list()
    for root, _, basename_list in os.walk(SRC):
        if os.path.basename(root) == "__pycache__":
            pyc_folder_list.append(root)

    for folder in pyc_folder_list:
        shutil.rmtree(folder)
    print("    all *.pyc file has been removed.")

    # install this package to all python version
    print("Uninstall %s from %s ..." % (PKG_NAME, DST))
    try:
        shutil.rmtree(DST)
        print("    Successfully uninstall %s" % PKG_NAME)
    except Exception as e:
        print("    %s" % e)

    print("Install %s to %s ..." % (PKG_NAME, DST))
    shutil.copytree(SRC, DST)
    print("    Complete!")


if __name__ == "__main__":
    install()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module provide advance csv I/O interface.
"""

from collections import OrderedDict
import pandas as pd
try:
    from .transform import itertuple
except:
    from pandas_mate.transform import itertuple


def iter_tuple_from_csv(path,
                        iterator=False,
                        chunksize=None,
                        skiprows=None,
                        nrows=None,
                        **kwargs):
    """A high performance, low memory usage csv file row iterator function.

    :param path: csv file path.
    :param iterator:
    :param chunksize:
    :param skiprows:
    :param nrows:

    :yield tuple: 

    **中文文档**

    对dataframe进行tuple风格的高性能行遍历。

    对用pandas从csv文件读取的dataframe进行逐行遍历时, iterrows和itertuple
    都不是性能最高的方法。这是因为iterrows要生成Series对象, 而itertuple
    也要对index进行访问。所以本方法是使用内建zip方法对所有的column进行打包
    解压, 所以性能上是最佳的。
    """
    kwargs["iterator"] = iterator
    kwargs["chunksize"] = chunksize
    kwargs["skiprows"] = skiprows
    kwargs["nrows"] = nrows

    if iterator is True:
        for df in pd.read_csv(path, **kwargs):
            for tp in itertuple(df):
                yield tp
    else:
        df = pd.read_csv(path, **kwargs)
        for tp in itertuple(df):
            yield tp


def index_row_dict_from_csv(path,
                            index_col=None,
                            iterator=False,
                            chunksize=None,
                            skiprows=None,
                            nrows=None,
                            use_ordered_dict=True,
                            **kwargs):
    """Read the csv into a dictionary. The key is it's index, the value
    is the dictionary form of the row.

    :param path: csv file path.
    :param index_col: None or str, the column that used as index.
    :param iterator:
    :param chunksize:
    :param skiprows:
    :param nrows:
    :param use_ordered_dict:

    :returns: {index_1: row1, index2: row2, ...} 

    **中文文档**

    读取csv, 选择一值完全不重复, 可作为index的列作为index, 生成一个字典
    数据结构, 使得可以通过index直接访问row。
    """
    _kwargs = dict(list(kwargs.items()))
    _kwargs["iterator"] = None
    _kwargs["chunksize"] = None
    _kwargs["skiprows"] = 0
    _kwargs["nrows"] = 1

    df = pd.read_csv(path, index_col=index_col, **_kwargs)
    columns = df.columns

    if index_col is None:
        raise Exception("please give index_col!")

    if use_ordered_dict:
        table = OrderedDict()
    else:
        table = dict()

    kwargs["iterator"] = iterator
    kwargs["chunksize"] = chunksize
    kwargs["skiprows"] = skiprows
    kwargs["nrows"] = nrows

    if iterator is True:
        for df in pd.read_csv(path, index_col=index_col, **kwargs):
            for ind, tp in zip(df.index, itertuple(df)):
                table[ind] = dict(zip(columns, tp))
    else:
        df = pd.read_csv(path, index_col=index_col, **kwargs)
        for ind, tp in zip(df.index, itertuple(df)):
            table[ind] = dict(zip(columns, tp))

    return table



#!/usr/bin/env python
# -*- coding: utf-8 -*-


def grouper_list(l, n):
    """Evenly divide list into fixed-length piece, no filled value if chunk
    size smaller than fixed-length.

    Example::

        >>> list(grouper(range(10), n=3)
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

    **中文文档**

    将一个列表按照尺寸n, 依次打包输出, 有多少输出多少, 并不强制填充包的大小到n。

    下列实现是按照性能从高到低进行排列的:

    - 方法1: 建立一个counter, 在向chunk中添加元素时, 同时将counter与n比较, 如果一致
      则yield。然后在最后将剩余的item视情况yield。
    - 方法2: 建立一个list, 每次添加一个元素, 并检查size。
    - 方法3: 调用grouper()函数, 然后对里面的None元素进行清理。
    """
    chunk = list()
    counter = 0
    for item in l:
        counter += 1
        chunk.append(item)
        if counter == n:
            yield chunk
            chunk = list()
            counter = 0
    if len(chunk) > 0:
        yield chunk



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.0.4"
__short_description__ = "A library extend pymongo module, makes CRUD easier, and more."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__maintainer__ = "Sanhe Hu"
__maintainer_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"


try:
    from .crud.insert import *
    from .crud.select import *
    from .crud.update import *
    from .query_builder import *
except ImportError:  # pragma: no cover
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module helps user to build pymongo query dict.
"""

import re


class Logic(object):
    @staticmethod
    def and_(*filters):
        return {"$and": list(filters)}

    @staticmethod
    def or_(*filters):
        return {"$or": list(filters)}

    @staticmethod
    def nor(*filters):
        return {"$nor": list(filters)}

    @staticmethod
    def not_(expression):
        return {"$not": expression}


class Comparison(object):
    """``==``, ``!=``, ``>``, ``>=``, ``<``, ``<=``.
    """
    @staticmethod
    def less_than_equal(value):
        """
        ``<=``

        Example::

            filters = {field: Comparison.less_than_equal(10)}
        """
        return {"$lte": value}

    @staticmethod
    def less_than(value):
        """
        ``<``

        Example::

            filters = {field: Comparison.less_than(10)}
        """
        return {"$lt": value}

    @staticmethod
    def greater_than_equal(value):
        """
        ``>=``

        Example::

            filters = {field: Comparison.greater_than_equal(10)}
        """
        return {"$gte": value}

    @staticmethod
    def greater_than(value):
        """
        ``>``

        Example::

            filters = {field: Comparison.greater_than(10)}
        """
        return {"$gt": value}

    @staticmethod
    def equal_to(value):
        """
        ``==``

        Example::

            filters = {field: Comparison.equal_to(10)}
        """
        return {"$eq": value}

    @staticmethod
    def not_equal_to(value):
        """
        ``!=``

        Example::

            filters = {field: Comparison.not_equal_to(10)}
        """
        return {"$ne": value}


class Lang(object):
    """
    Language code Collection.
    """
    English = "en"
    French = "fr"
    German = "de"
    Italian = "it"
    Portuguese = "pt"
    Russian = "ru"
    Spanish = "es"
    SimplifiedChineses = "zhs"
    TraditionalChineses = "zht"


class Text(object):
    """Search text by:

    - startswith
    - endswith
    - contain sub string
    - full text serach
    """
    @staticmethod
    def startswith(text, ignore_case=True):
        """
        Test if a string-field start with ``text``.

        Example::

            filters = {"path": Text.startswith(r"C:\\")}
        """
        if ignore_case:
            compiled = re.compile(
                "^%s" % text.replace("\\", "\\\\"), re.IGNORECASE)
        else:  # pragma: no cover
            compiled = re.compile("^%s" % text.replace("\\", "\\\\"))

        return {"$regex": compiled}

    @staticmethod
    def endswith(text, ignore_case=True):
        """
        Test if a string-field end with ``text``.

        Example::

            filters = {"path": Text.endswith(".exe")}
        """
        if ignore_case:
            compiled = re.compile(
                "%s$" % text.replace("\\", "\\\\"), re.IGNORECASE)
        else:  # pragma: no cover
            compiled = re.compile("%s$" % text.replace("\\", "\\\\"))

        return {"$regex": compiled}

    @staticmethod
    def contains(text, ignore_case=True):
        """
        Test if a string-field has substring of ``text``.

        Example::

            filters = {"path": Text.contains("pymongo_mate")}
        """
        if ignore_case:
            compiled = re.compile(
                "%s" % text.replace("\\", "\\\\"), re.IGNORECASE)
        else:  # pragma: no cover
            compiled = re.compile("%s" % text.replace("\\", "\\\\"))

        return {"$regex": compiled}

    @staticmethod
    def fulltext(search, lang=Lang.English, ignore_case=True):
        """Full text search.

        Example::

            filters = Text.fulltext("python pymongo_mate")

        .. note::

            This field doesn't need to specify field.
        """
        return {
            "$text": {
                "$search": search,
                "$language": lang,
                "$caseSensitive": not ignore_case,
                "$diacriticSensitive": False,
            }
        }


class Array(object):
    """Array query builder.
    """
    @staticmethod
    def element_match(filters):
        """
        Test if any of items match the criterion.

        Example::

            data = [
                {"_id": 1, "results": [ 82, 85, 88 ]},
                {"_id": 2, "results": [ 75, 88, 89 ]},
            ]

            filters = {"results": {"$elemMatch": {"$gte": 80, "$lt": 85 }}}

            # equals to
            filters = {
                "results": Array.element_match({"$gte": 80, })
            }
        """
        return {"$elemMatch": filters}

    @staticmethod
    def include_all(items):
        """
        Test if an array-like field include all these items.

        Example::

            {"tag": ["a", "b", "c"]} include all ["a", "c"]
        """
        return {"$all": items}

    @staticmethod
    def include_any(items):
        """
        Test if an array-like field include any of these items.

        Example::

            {"tag": ["a", "b", "c"]} include any of ["c", "d"]
        """
        return {"$in": items}

    @staticmethod
    def exclude_all(items):
        """
        Test if an array-like field doesn't include any of these items.

        Example::

            {"tag": ["a", "b", "c"]} doesn't include any of ["d", "e"]
        """
        return {"$nin": items}

    @staticmethod
    def exclude_any(items):
        """
        Test if an array-like field doesn't include at lease one of
        these items.

        Example::

            {"tag": ["a", "b", "c"]} doesn't include "d" from ["c", "d"]
        """
        return {"$not": {"$all": items}}

    @staticmethod
    def item_in(items):
        """
        Single item is in item sets.

        Example::

            {"item_type": "Fruit"}, "Fruit" in ["Fruit", "Meat"]
        """
        return {"$in": items}

    @staticmethod
    def item_not_in(items):
        """
        Single item is not in item sets.

        Example::

            {"item_type": "Seafood"}, "Fruit" not in ["Fruit", "Meat"]
        """
        return {"$nin": items}

    @staticmethod
    def size(n):
        """
        Test if size of an array-like field is n.

        Example::

            filters = {"cart.items": Array.size(3)}
        """
        return {"$size": n}


class Geo2DSphere(object):
    """Geosphere query builder.
    """
    @staticmethod
    def near(lat, lng, max_dist=None, unit_miles=False):
        """Find document near a point.

        For example:: find all document with in 25 miles radius from 32.0, -73.0.
        """
        filters = {
            "$nearSphere": {
                "$geometry": {
                    "type": "Point",
                    "coordinates": [lng, lat],
                }
            }
        }
        if max_dist:
            if unit_miles:  # pragma: no cover
                max_dist = max_dist / 1609.344
            filters["$nearSphere"]["$maxDistance"] = max_dist
        return filters


def exists(true_or_false):
    """Test if a field exists.
    """
    return {"$exists": true_or_false}


is_exists = {"$exists": True}
"""Match a field exists.
"""

is_not_exists = {"$exists": False}
"""Match a field not exists.
"""


def mod(divisor, remainder):
    """Test mod(<divisor>) is <remainder>.
    """
    return {"$mod": [divisor, remainder]}


class TypeCode(object):
    """MongoDB Bson type code.
    """
    Double = 1
    String = 2
    Object = 3
    Array = 4
    Binary_Data = 5
    Undefined = 6  # Deprecated
    ObjectId = 7
    Boolean = 8
    Date = 9
    Null = 10
    Regular_Expression = 11
    DBPointer = 12  # Deprecated
    JavaScript = 13
    Symbol = 14  # Deprecated
    JavaScript_with_Scope = 15
    Integer32 = 16
    TimeStamp = 17
    Integer64 = 18
    Decimal128 = 19
    MinKey = -1
    MaxKey = 127
    Number = "number"


def type_is(type_code):
    """Test field type is certain type.
    """
    return {"$type": type_code}


class PipeLine(object):
    @staticmethod
    def random_sample(filters=None, n=5):
        pipeline = list()
        if filters is not None:
            pipeline.append({"$match": filters})
        pipeline.append({"$sample": {"size": n}})
        return pipeline



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import pymongo
import mongomock

py_ver = "%s%s" % (sys.version_info.major, sys.version_info.minor)
col_name = "test_col_py%s" % py_ver

try:
    # Create Client
    # local
    #     client = pymongo.MongoClient(
    #         "localhost", 27017, serverSelectionTimeoutMS=1000)

    # mlab cloud
    url = "mongodb://{dbuser}:{dbpassword}@ds113063.mlab.com:13063/devtest".\
        format(dbuser="admin", dbpassword="&2z7#tMH6BJt")
    client = pymongo.MongoClient(url, serverSelectionTimeoutMS=1000)

    # Use Database
    db = client.__getattr__("devtest")

    # Connect to Collection
    col_real = db.__getattr__(col_name)
    col_real.drop()
except Exception as e:
    col_real = None

# Create Mongomock Client
col_mock = mongomock.MongoClient().devtest.__getattr__(col_name)
col_mock.drop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
import pymongo
try:
    from ..utils import grouper_list
except:  # pragma: no cover
    from pymongo_mate.utils import grouper_list

try:
    try:
        from .pkg.pandas_mate import transform
    except:
        from pymongo_mate.pkg.pandas_mate import transform
except:  # pragma: no cover
    pass


__all__ = [
    "smart_insert",
    "insert_data_frame",
]


def smart_insert(col, data, minimal_size=5):
    """An optimized Insert strategy.

    **中文文档**

    在Insert中, 如果已经预知不会出现IntegrityError, 那么使用Bulk Insert的速度要
    远远快于逐条Insert。而如果无法预知, 那么我们采用如下策略:

    1. 尝试Bulk Insert, Bulk Insert由于在结束前不Commit, 所以速度很快。
    2. 如果失败了, 那么对数据的条数开平方根, 进行分包, 然后对每个包重复该逻辑。
    3. 若还是尝试失败, 则继续分包, 当分包的大小小于一定数量时, 则使用逐条插入。
      直到成功为止。

    该Insert策略在内存上需要额外的 sqrt(nbytes) 的开销, 跟原数据相比体积很小。
    但时间上是各种情况下平均最优的。
    """
    if isinstance(data, list):
        # 首先进行尝试bulk insert
        try:
            col.insert(data)
        # 失败了
        except pymongo.errors.DuplicateKeyError:
            # 分析数据量
            n = len(data)
            # 如果数据条数多于一定数量
            if n >= minimal_size ** 2:
                # 则进行分包
                n_chunk = math.floor(math.sqrt(n))
                for chunk in grouper_list(data, n_chunk):
                    smart_insert(col, chunk, minimal_size)
            # 否则则一条条地逐条插入
            else:
                for doc in data:
                    try:
                        col.insert(doc)
                    except pymongo.errors.DuplicateKeyError:
                        pass
    else:  # pragma: no cover
        try:
            col.insert(data)
        except pymongo.errors.DuplicateKeyError:
            pass


def insert_data_frame(col, df, int_col=None, binary_col=None, minimal_size=5):
    """Insert ``pandas.DataFrame``.

    :param col: :class:`pymongo.collection.Collection` instance.
    :param df: :class:`pandas.DataFrame` instance.
    :param int_col: list of integer-type column.
    :param binary_col: list of binary-type column.
    """
    data = transform.to_dict_list_generic_type(df,
                                               int_col=int_col,
                                               binary_col=binary_col)
    smart_insert(col, data, minimal_size)



from __future__ import print_function
import argparse
import sys

from .factory import from_csv
from ._compact import StringIO


def main():
    text_in = sys.stdin.read()
    if text_in:
        print(from_csv(StringIO.StringIO(text_in)))
        return

    parser = argparse.ArgumentParser(description='A simple Python library designed to make it quick and easy to '
                                     'represent tabular data in visually appealing ASCII tables.')
    parser.add_argument('--csv', help='CSV file name')
    args = parser.parse_args()
    with open(args.csv) as fp:
        print(from_csv(fp))


if __name__ == '__main__':
    main()



import sys


py3k = sys.version_info[0] >= 3

if py3k:
    str_types = (str, )
    unicode_ = str
    basestring_ = str
    itermap = map
    iterzip = zip
    uni_chr = chr
    from html.parser import HTMLParser
    import io as StringIO
else:
    unicode_ = unicode
    basestring_ = basestring
    str_types = (unicode, str)
    import itertools
    itermap = itertools.imap
    iterzip = itertools.izip
    uni_chr = unichr
    from HTMLParser import HTMLParser
    import StringIO

HTMLParser, StringIO


if py3k and sys.version_info[1] >= 2:
    from html import escape
else:
    from cgi import escape
escape
